/*
 * Copyright (c) 2004 Anil Madhavapeddy <anil@recoil.org>
 * Copyright (c) 2004 Intel Research Cambridge
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $Id: blueproxy.c,v 1.31 2004/09/07 15:53:32 avsm Exp $
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <sys/poll.h>
#include <sys/stat.h>
#include <pthread.h>
#include <getopt.h>

#include "atomicio.h"
#include "log.h"
#include "blueproxy.h"

#define DEFAULT_SOCKPATH "/tmp/bluespot"
#define DEFAULT_PACKETSIZE 128
#define MAX_PACKETSIZE 65535
#define MAX_CHANNEL 16

char hostname[1024] = "localhost";
char search_name[256] = "Serial Port";
u_int16_t port = 20000;
char bbid[1024] = "bbid unknown\n";
int got_id = 0;
int enquiry = 0;
char interface[64] = DEFAULT_INTERFACE;
size_t packet_size = DEFAULT_PACKETSIZE;
int channels[] = { 1 };
size_t num_channels = 1;

/* Prototypes */
struct sockpair {
    int from;
    int to;
    int channel;
    char id[128];
};

static int connect_tcp_socket(char *);
static void *splice_sockets(void *);
static void *listen_on_channel(void *);
static pthread_t *connect_to_remote(char *, int);
static ssize_t transfer_data(char *, char *, int, int);
static void free_service_results(struct service_results *);
static void usage(void);

/* Party on Wayne */
int
main(int argc, char **argv)
{
    int c, client, rchan;
    pthread_t *tids;
    extern char *optarg;
    extern int optind, optopt;

    init_log();
    client = 0;
    rchan = 0;
    set_log_level(L_DEBUG);

    while ((c = getopt(argc, argv, "n:r:e:c:b:p:h:i:?")) != -1) {
        int i;
        switch (c) {
        case 'e':
            enquiry = 1;
            snprintf(search_name, sizeof search_name, "%s", optarg);
            break;
        case 'i':
            strncpy(interface, optarg, sizeof(interface) - 1);
            interface[sizeof(interface) - 1] = '\0';
            break;
        case 'h':
            strncpy(hostname, optarg, sizeof(hostname) - 1);
            hostname[sizeof(hostname) - 1] = '\0';
            break;
        case 'b':
            i = atoi(optarg);
            if (i > 0 && i < MAX_PACKETSIZE)
                packet_size = i;
            else
                logmsg(NULL, L_INFO, "Invalid packet size %s", optarg);
            break;
        case 'p':
            i = atoi(optarg);
            if (i > 0 && i < 65535)
                port = (u_int16_t) i;
            else
                logmsg(NULL, L_INFO, "Invalid port %s", optarg);
            break;
        case 'r':
            client = 1;
            strncpy(interface, optarg, sizeof(interface) - 1);
            interface[sizeof(interface) - 1] = '\0';
            break;
        case 'c':
            rchan = atoi(optarg);
            if (rchan < 0 || rchan > MAX_CHANNEL)
                errx(1, "invalid channel %d", rchan);
            break;
        case 'n':
            snprintf(bbid, sizeof bbid, "bbid %s\n", optarg);
            got_id = 1;
            break;
        case '?':
        default:
            usage();
        }
    }

    logmsg(NULL, L_DEBUG, "interface: %s", interface);
    logmsg(NULL, L_DEBUG, "socket path: %s", hostname);
    logmsg(NULL, L_DEBUG, "packet size: %d", packet_size);

    if (enquiry) {
        while (1) {
            struct service_results *res;
            logmsg("scan", L_DEBUG, "looping to scan");
            res = doBluetoothServiceInquiry("scan", search_name);
            while (res) {
                pthread_t *tid;
                tid = connect_to_remote(res->bdaddr, res->channel);
                free(tid);
                res = res->next;
            }
            free_service_results(res);
            sleep(2);
        }
    } else if (client) {
        pthread_t *tid;
        tid = connect_to_remote(interface, rchan);
        pthread_join(*tid, NULL);
    } else {
        size_t i;
        tids = malloc(sizeof(pthread_t) * num_channels);
        if (!tids)
            err(1, "tids malloc");

        for (i = 0; i < num_channels; i++) {
            int ret;
            int *chan = malloc(sizeof(int));
            if (chan == NULL)
                err(1, "malloc");
            *chan = channels[i];
            ret =
                pthread_create(&tids[i], NULL, listen_on_channel,
                               (void *) chan);
            if (ret < 0)
                err(1, "pthread_create");
        }

        for (i = 0; i < num_channels; i++)
            pthread_join(tids[i], NULL);
    }

    return 0;
}

/* Display usage information */
static void
usage(void)
{
    fprintf(stderr, "Blueproxy v1.1 usage:\n");
    fprintf(stderr, "Listen mode: blueproxy <options>\n");
    fprintf(stderr, "Splice mode: blueproxy -r remote_bdaddr -c channel <options>\n");
    fprintf(stderr, "Scan mode  : blueproxy -e service_name <options>\n");
    fprintf(stderr, "  options:  [-h host] [-p port] [-i interface] [-b packet size] [-n preamble]\n");
    exit(1);
}

/* Free a linked list of service results from a service scan */
static void
free_service_results(struct service_results *res)
{
    while (res) {
        struct service_results *old = res;
        res = res->next;
        free(old);
    }
}

/* Connect to remote Bluetooth host and splice the connections
   together */
static pthread_t *
connect_to_remote(char *iface, int channel)
{
    int s = -1, r = -1;
    struct sockpair *socks = NULL;
    char id[32];
    pthread_t *tid;
    snprintf(id, sizeof id, "ch #%d", channel);
    logmsg(id, L_INFO, "Connecting to %s (chan %d)", iface, channel);

    s = connectBluetoothSocket(id, iface, channel);
    if (s < 0) {
        logmsg(id, L_CRITICAL, "Error connecting bluetooth: %s",
               strerror(errno));
        goto cleanup;
    }
    r = connect_tcp_socket(id);
    if (r < 0) {
        logmsg(id, L_CRITICAL, "tcp socket: %s", strerror(errno));
        goto cleanup;
    }
    socks = malloc(sizeof(struct sockpair));
    if (!socks) {
        logmsg(id, L_CRITICAL, "malloc failed: %s", strerror(errno));
        goto cleanup;
    }
    socks->from = s;
    socks->to = r;
    socks->channel = channel;
    snprintf(socks->id, sizeof(socks->id), "ch# %d", channel);

    if (got_id) {
        logmsg(id, L_DEBUG, "sending bbid %s", bbid);
        if (atomicio(awrite, r, bbid, strlen(bbid)) <= 0) {
            logmsg(id, L_CRITICAL, "error sending bbid: %s",
                   strerror(errno));
            goto cleanup;
        }
    }

    tid = (pthread_t *) malloc(sizeof(pthread_t));
    if (!tid) {
        logmsg(id, L_CRITICAL, "malloc failed: %s", strerror(errno));
        goto cleanup;
    }
    if (pthread_create(tid, NULL, splice_sockets, (void *) socks) < 0) {
        logmsg(id, L_CRITICAL, "pthread_create: %s", strerror(errno));
        goto cleanup;
    }
    return tid;

cleanup:
    if (s >= 0)
        close(s);
    if (r >= 0)
        close(r);
    return NULL;
}

/* Listen on a channel, spawning new threads when connections
 * come in until a hard error condition is hit */
static void *
listen_on_channel(void *chan)
{
    int channel = *((int *) chan);
    int s = -1, a = -1, r = -1;
    char id[32];
    struct sockpair *socks = NULL;

    snprintf(id, sizeof id, "ch #%d", channel);
    logmsg(NULL, L_INFO, "Listening on channel %d", channel);

    registerBlueproxyService(id, interface, channel);
    logmsg(NULL, L_INFO, "Registered Serial Port service on channel %d",
           channel);

    s = createBluetoothSocket(id, interface, channel);
    if (s < 0) {
        logmsg(id, L_CRITICAL, "error creating bluetooth socket: %s",
               strerror(errno));
        goto cleanup;
    }

    while (1) {
        pthread_t tid;
        int ret;
        a = acceptBluetoothSocket(id, s);
        if (a < 0) {
            logmsg(id, L_CRITICAL, "accept bluetooth socket: %s",
                   strerror(errno));
            goto cleanup;
        }
        r = connect_tcp_socket(id);
        if (r < 0) {
            logmsg(id, L_CRITICAL, "tcp socket: %s", strerror(errno));
            if (a >= 0)
                close(a);
            continue;
        }
        socks = malloc(sizeof(struct sockpair));
        if (!socks) {
            logmsg(id, L_CRITICAL, "malloc failed: %s", strerror(errno));
            goto cleanup;
        }
        socks->from = a;
        socks->to = r;
        socks->channel = channel;
        snprintf(socks->id, sizeof(socks->id), "ch# %d", channel);
        fprintf(stderr, "here2\n");
        ret = pthread_create(&tid, NULL, splice_sockets, (void *) socks);
        if (ret < 0) {
            logmsg(id, L_CRITICAL, "pthread_create: %s", strerror(errno));
            goto cleanup;
        } else
            pthread_join(tid, NULL);
    }
cleanup:
    free(chan);
    if (s >= 0)
        close(s);
    if (a >= 0)
        close(a);
    if (r >= 0)
        close(r);
    return NULL;
}

/* Attempt to connect to a tcp socket */
static int
connect_tcp_socket(char *id)
{
    int s;
    struct sockaddr_in sa;
    struct hostent *hp;

    if ((hp = gethostbyname(hostname)) == NULL) {
        logmsg(id, L_CRITICAL, "Unable to resolve %s", hostname);
        return -1;
    }

    logmsg(id, L_DEBUG, "Connecting tcp socket to %s:%d", hostname, port);
    s = socket(AF_INET, SOCK_STREAM, 0);
    if (s < 0)
        return s;

    memset(&sa, 0, sizeof sa);
    sa.sin_family = AF_INET;
    bcopy(hp->h_addr, &sa.sin_addr, hp->h_length);
    sa.sin_port = htons(port);

    if (connect(s, (struct sockaddr *) &sa, sizeof sa) < 0)
        return -1;

    logmsg(id, L_INFO, "Connection to %s:%d established", hostname, port);
    return s;
}

/* Read from socket and write it out to second socket. */
static ssize_t
transfer_data(char *id, char *buf, int from, int to)
{
    ssize_t x, y;
    x = atomicio(read, from, buf, packet_size);
    if (x <= 0) {
        logmsg(id, L_INFO, "read err (%d): %s", x, strerror(errno));
        return -1;
    }
    y = atomicio(awrite, to, buf, x);
    if (y != x) {
        logmsg(id, L_INFO, "write err (%d): %s", y, strerror(errno));
        return -1;
    }
    return y;
}

/* Thread which reads from one socket and writes to another
 * until it hits an error condition */
static void *
splice_sockets(void *data)
{
    struct sockpair *sp = (struct sockpair *) data;
    struct pollfd rfds[2];
    int ret;
    ssize_t x;
    char bttopc[64];
    char pctobt[64];
    char *buf;

    snprintf(bttopc, sizeof bttopc, "%s bt->pc", sp->id);
    snprintf(pctobt, sizeof pctobt, "%s pc->bt", sp->id);

    buf = malloc(packet_size);
    if (!buf) {
        logmsg(sp->id, L_CRITICAL, "malloc failed: %s", strerror(errno));
        goto cleanup;
    }

    logmsg(sp->id, L_DEBUG, "splice_sockets started");
    rfds[0].fd = sp->from;
    rfds[0].events = POLLIN;
    rfds[0].revents = 0;
    rfds[1].fd = sp->to;
    rfds[1].events = POLLIN;
    rfds[1].revents = 0;
    while ((ret = poll(rfds, 2, -1))) {
        if (ret < 0) {
            logmsg(sp->id, L_INFO, "poll error: %s", strerror(errno));
            goto cleanup;
        }
        memset(buf, 0, packet_size);
        /* check for any error conditions and exit if found */
        if (rfds[0].revents & POLLHUP || rfds[0].revents & POLLERR ||
                rfds[0].revents & POLLNVAL) {
            logmsg(bttopc, L_INFO, "BT socket closed");
            goto cleanup;
        }
        if (rfds[1].revents & POLLHUP || rfds[1].revents & POLLERR ||
                rfds[1].revents & POLLNVAL) {
            logmsg(pctobt, L_INFO, "PC socket closed");
            goto cleanup;
        }

        /* check for outstanding data and transfer if its present */
        if (rfds[0].revents & POLLIN) {
            x = transfer_data(bttopc, buf, sp->from, sp->to);
            if (x < 0)
                goto cleanup;
            logmsg(bttopc, L_DEBUG, "read/wrote %d bytes:\n%s", x, buf);
        }
        if (rfds[1].revents & POLLIN) {
            x = transfer_data(pctobt, buf, sp->to, sp->from);
            if (x < 0)
                goto cleanup;
            logmsg(pctobt, L_DEBUG, "read/wrote %d bytes:\n%s", x, buf);
        }
        rfds[0].revents = 0;
        rfds[1].revents = 0;
    }

cleanup:
    logmsg(sp->id, L_INFO, "Spliced connection closed");
    close(sp->from);
    close(sp->to);
    free(buf);
    free(data);
    return NULL;
}
