/*
 * Copyright (c) 2004 Anil Madhavapeddy <anil@recoil.org>
 * Copyright (c) 2004 Intel Research Cambridge
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $Id: freebsd.c,v 1.8 2004/09/07 15:52:04 avsm Exp $
 */

#include <stdio.h>
#include <string.h>
#include <bluetooth.h>
#include <sdp.h>
#include "blueproxy.h"
#include "log.h"

static void *sdp_sess = NULL;

/* Convert interface string into a bdaddr_t */
static void
bdaddr_of_iface_name(char *iface_name, bdaddr_t * addr) {
    if (!bt_aton(iface_name, addr))
        memcpy(addr, NG_HCI_BDADDR_ANY, sizeof addr);
}

/* create a ng_bluetooth listening socket */
int
createBluetoothSocket(char *id, char *iface_name, int channel) {
    struct sockaddr_rfcomm sock_addr;
    bdaddr_t addr;
    int s;

    bdaddr_of_iface_name(iface_name, &addr);

    s = socket(PF_BLUETOOTH, SOCK_STREAM, BLUETOOTH_PROTO_RFCOMM);
    if (s < 0)
        return s;
    else
        logmsg(id, L_DEBUG, "Created ng_bluetooth socket");

    logmsg(id, L_DEBUG, "Binding to interface %s", bt_ntoa(&addr, NULL));

    memset(&sock_addr, 0, sizeof sock_addr);
    sock_addr.rfcomm_len = sizeof sock_addr;
    sock_addr.rfcomm_family = AF_BLUETOOTH;
    memcpy(&sock_addr.rfcomm_bdaddr, &addr,
           sizeof(sock_addr.rfcomm_bdaddr));
    sock_addr.rfcomm_channel = channel;

    if (bind(s, (struct sockaddr *) &sock_addr, sizeof(sock_addr)) < 0)
        return -1;
    else
        logmsg(id, L_DEBUG, "Done ng_bluetooth bind");

    if (listen(s, 10) < 0)
        return -1;
    else
        logmsg(id, L_DEBUG, "Done ng_bluetooth listen");

    return s;
}

/* accept a connection on a freebsd bluetooth socket */
int
acceptBluetoothSocket(char *id, int s) {
    struct sockaddr_rfcomm r_addr;
    socklen_t opt = sizeof r_addr;
    int a;

    logmsg(id, L_DEBUG, "Starting ng_bluetooth accept");
    a = accept(s, (struct sockaddr *) &r_addr, &opt);
    if (a < 0)
        return a;

    logmsg(id, L_INFO, "Accepted connection from: %s",
           bt_ntoa(&r_addr.rfcomm_bdaddr, NULL));

    return a;
}

/* connect a socket to a remote bluetooth host */
int
connectBluetoothSocket(char *id, char *remote, int channel) {
    errx(1, "Not implemented on FreeBSD yet");
}

/* Do a service inquiry looking for Bluetooth devices that offer
 * a particular type of serial port service specified in search_name */
struct service_results *
    doBluetoothServiceInquiry(char *id, char *search_name) {
    errx(1, "Not implemented on FreeBSD yet");
}

/* Register the Serial Port profile for the listening channel */
void
registerBlueproxyService(char *id, char *iface_name, int channel) {
    sdp_sp_profile_t sp;
    bdaddr_t addr;

    bdaddr_of_iface_name(iface_name, &addr);

    /* Connect to SDP server if we havent done so yet */
    if (sdp_sess == NULL) {
        sdp_sess = sdp_open_local(NULL);
        if (sdp_sess == NULL) {
            logmsg(id, L_CRITICAL, "Unable to create local SDP session");
            exit(1);
        }

        if (sdp_error(sdp_sess) != 0) {
            logmsg(id, L_CRITICAL, "Unable to open local SDP session: %s",
                   strerror(sdp_error(sdp_sess)));
            exit(1);
        }
    }

    memset(&sp, 0, sizeof sp);
    sp.server_channel = channel;
    if (sdp_register_service(sdp_sess, SDP_SERVICE_CLASS_SERIAL_PORT,
                             &addr, (void *) &sp, sizeof sp, NULL) != 0) {
        logmsg(id, L_CRITICAL,
               "Unable to register Serial Port Service: %s",
               strerror(sdp_error(sdp_sess)));
        exit(1);
    }
}

/* Call on program exit to unregister the added services cleanly */
void
unregisterBlueproxyService() {
    if (sdp_sess != NULL)
        sdp_close(sdp_sess);
}
