/*
 * Copyright (c) 2004 Anil Madhavapeddy <anil@recoil.org>
 * Copyright (c) 2004 Intel Research Cambridge
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $Id: linux_bluez.c,v 1.18 2004/09/07 15:52:04 avsm Exp $
 */

#include "blueproxy.h"
#include "log.h"
#include <bluetooth/bluetooth.h>
#include <bluetooth/rfcomm.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>

static sdp_session_t *sdp_sess = NULL;

/* create a bluez listening socket */
int
createBluetoothSocket(char *id, char *iface_name, int channel) {
    struct sockaddr_rc local_addr;
    char actual_interface[64];
    bdaddr_t interface;
    int s;

    s = socket(AF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM);
    if (s < 0)
        return s;
    else
        logmsg(id, L_DEBUG, "Created bluez socket");

    if (str2ba(iface_name, &interface)) {
        logmsg(id, L_CRITICAL, "Invalid interface: %s", iface_name);
        return -1;
    }

    ba2str(&interface, actual_interface);
    logmsg(id, L_DEBUG, "Binding to interface %s", actual_interface);

    memset(&local_addr, 0, sizeof local_addr);
    local_addr.rc_family = AF_BLUETOOTH;
    bacpy(&local_addr.rc_bdaddr, &interface);
    local_addr.rc_channel = htobs(channel);

    if (bind(s, (struct sockaddr *) &local_addr, sizeof(local_addr)) < 0)
        return -1;
    else
        logmsg(id, L_DEBUG, "Done bluez bind");

    if (listen(s, 10) < 0)
        return -1;
    else
        logmsg(id, L_DEBUG, "Done bluez listen");

    return s;
}

/* accept a connection on a bluez socket */
int
acceptBluetoothSocket(char *id, int s) {
    struct sockaddr_rc r_addr;
    socklen_t opt = sizeof r_addr;
    int a;

    logmsg(id, L_DEBUG, "Starting bluez accept");
    a = accept(s, (struct sockaddr *) &r_addr, &opt);
    if (a < 0)
        return a;

    logmsg(id, L_INFO, "Accepted connection from: %s",
           batostr(&r_addr.rc_bdaddr));

    return a;
}

/* connect bluez socket to remote host */
int
connectBluetoothSocket(char *id, char *remote, int channel) {
    struct sockaddr_rc rem_addr, loc_addr;
    bdaddr_t target;

    int s = socket(PF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM);
    if (s < 0)
        return s;

    memset(&loc_addr, 0, sizeof loc_addr);
    loc_addr.rc_family = AF_BLUETOOTH;
    bacpy(&loc_addr.rc_bdaddr, BDADDR_ANY);
    if (bind(s, (struct sockaddr *) &loc_addr, sizeof loc_addr) < 0)
        return s;

    if (str2ba(remote, &target)) {
        logmsg(id, L_CRITICAL, "Invalid remote address %s", remote);
        return -1;
    }

    memset(&rem_addr, 0, sizeof rem_addr);
    rem_addr.rc_family = AF_BLUETOOTH;
    bacpy(&rem_addr.rc_bdaddr, &target);
    rem_addr.rc_channel = htobs(channel);

    if (connect(s, (struct sockaddr *) &rem_addr, sizeof rem_addr) < 0) {
        logmsg(id, L_CRITICAL, "Error in connection: %s", strerror(errno));
        return -1;
    } else
        logmsg(id, L_DEBUG, "connected to remote host %s on channel %d",
               remote, channel);
    return s;
}

/* Iterate through a set of service descriptors and look out for the
 * RFCOMM channel number for this protocol */
static void
getServiceDescriptor(void *value, void *channel) {
    sdp_data_t *data = (sdp_data_t *) value;
    int *channel_i = (int *) channel;
    int protocol = 0;
    for (; data; data = data->next) {
        switch (data->dtd) {
        case SDP_UUID16:
        case SDP_UUID32:
        case SDP_UUID128:
            protocol = sdp_uuid_to_proto(&data->val.uuid);
            break;
        case SDP_UINT8:
            if (protocol == RFCOMM_UUID)
                *channel_i = data->val.uint8;
            break;
        }
    }
}

/* Just pass a protocol descriptor through to analyse the
 * server descriptors */
static void
getProtocolDescriptor(void *value, void *channel) {
    sdp_list_t *seq = (sdp_list_t *) value;
    sdp_list_foreach(seq, getServiceDescriptor, channel);
}

/* Do a service inquiry looking for Bluetooth devices that offer
 * a particular type of serial port service specified in search_name */
struct service_results *
    doBluetoothServiceInquiry(char *id, char *search_name) {
    u_int16_t service_type = SERIAL_PORT_SVCLASS_ID;
    inquiry_info ii[32];
    uuid_t group;
    uint32_t range = 0x0000ffff;
    sdp_list_t *attrid, *search, *seq, *next;
    u_int8_t count = 0;
    char nice_bdaddr[20];
    struct service_results *head = NULL;
    struct service_results *tail = NULL;
    int i;

    logmsg(id, L_INFO, "Starting general inquiry");

    if (sdp_general_inquiry(ii, sizeof ii, 8, &count) < 0) {
        logmsg(id, L_CRITICAL, "Inquiry failed");
        return NULL;
    }

    for (i = 0; i < count; i++) {
        sdp_session_t *session;
        bdaddr_t interface;
        bacpy(&interface, BDADDR_ANY);
        session =
            sdp_connect(&interface, &ii[i].bdaddr, SDP_RETRY_IF_BUSY);
        ba2str(&ii[i].bdaddr, nice_bdaddr);
        if (!session) {
            /* not a critical error - remote device doesnt want us to browse */
            logmsg(id, L_DEBUG, "Unable to connect to %s", nice_bdaddr);
            continue;
        }
        logmsg(id, L_DEBUG, "Found bdaddr %s", nice_bdaddr);
        attrid = sdp_list_append(0, &range);
        sdp_uuid16_create(&group, service_type);
        search = sdp_list_append(0, &group);
        if (sdp_service_search_attr_req
                (session, search, SDP_ATTR_REQ_RANGE, attrid, &seq)) {
            logmsg(id, L_CRITICAL, "Service search failed: %s",
                   strerror(errno));
            sdp_close(session);
        }
        sdp_list_free(attrid, 0);
        sdp_list_free(search, 0);

        for (; seq; seq = next) {
            sdp_list_t *list = NULL;
            sdp_record_t *record = (sdp_record_t *) seq->data;
            sdp_data_t *data =
                sdp_data_get(record, SDP_ATTR_SVCNAME_PRIMARY);
            if (data && !strcmp(data->val.str, search_name)) {
                logmsg(id, L_INFO, "Found %s at %s", search_name,
                       nice_bdaddr);
                if (!sdp_get_access_protos(record, &list)) {
                    struct service_results *r = NULL;
                    int channel = -1;
                    sdp_list_foreach(list, getProtocolDescriptor,
                                     &channel);
                    sdp_list_free(list, (sdp_free_func_t) sdp_data_free);
                    if (channel != -1) {
                        r = (struct service_results *) calloc(1,
                                                              sizeof(struct
                                                                     service_results));
                        snprintf(r->bdaddr, sizeof(r->bdaddr), "%s",
                                 nice_bdaddr);
                        r->channel = channel;
                        if (head == NULL) {
                            head = r;
                            tail = r;
                        } else {
                            tail->next = r;
                            tail = r;
                        }
                    }
                    logmsg(id, L_DEBUG, "channel ==== %d", channel);
                }
            }
            next = seq->next;
            free(seq);
            sdp_record_free(record);
        }
        sdp_close(session);
    }
    return head;
}

void
registerBlueproxyService(char *id, char *iface_name, int chan) {
    sdp_list_t *svclass_id, *apseq, *proto[2], *profiles, *root, *aproto;
    uuid_t root_uuid, sp_uuid, l2cap, rfcomm;
    sdp_profile_desc_t profile;
    sdp_record_t record;
    uint8_t chan_u8 = chan;
    sdp_data_t *channel;
    bdaddr_t iface;

    if (str2ba(iface_name, &iface)) {
        logmsg(id, L_CRITICAL, "Invalid interface: %s", iface_name);
        exit(1);
    }

    if (!sdp_sess) {
        sdp_sess = sdp_connect(&iface, BDADDR_LOCAL, SDP_RETRY_IF_BUSY);
        if (!sdp_sess) {
            logmsg(id, L_CRITICAL, "Couldnt connect to SDP daemon");
            exit(1);
        }
    }

    memset((void *) &record, 0, sizeof(sdp_record_t));
    record.handle = 0xffffffff;
    sdp_uuid16_create(&root_uuid, PUBLIC_BROWSE_GROUP);
    root = sdp_list_append(0, &root_uuid);
    sdp_set_browse_groups(&record, root);
    sdp_list_free(root, 0);

    sdp_uuid16_create(&sp_uuid, SERIAL_PORT_SVCLASS_ID);
    svclass_id = sdp_list_append(0, &sp_uuid);
    sdp_set_service_classes(&record, svclass_id);
    sdp_list_free(svclass_id, 0);

    sdp_uuid16_create(&profile.uuid, SERIAL_PORT_PROFILE_ID);
    profile.version = 0x0100;
    profiles = sdp_list_append(0, &profile);
    sdp_set_profile_descs(&record, profiles);
    sdp_list_free(profiles, 0);

    sdp_uuid16_create(&l2cap, L2CAP_UUID);
    proto[0] = sdp_list_append(0, &l2cap);
    apseq = sdp_list_append(0, proto[0]);

    sdp_uuid16_create(&rfcomm, RFCOMM_UUID);
    proto[1] = sdp_list_append(0, &rfcomm);
    channel = sdp_data_alloc(SDP_UINT8, &chan_u8);
    proto[1] = sdp_list_append(proto[1], channel);
    apseq = sdp_list_append(apseq, proto[1]);

    aproto = sdp_list_append(0, apseq);
    sdp_set_access_protos(&record, aproto);

    sdp_set_info_attr(&record, "Blueproxy Serial Port", 0, 0);

    if (sdp_record_register(sdp_sess, &record, 0)) {
        logmsg(id, L_CRITICAL, "Serial Port Service registration failed");
        exit(1);
    }

    sdp_data_free(channel);
    sdp_list_free(proto[0], 0);
    sdp_list_free(proto[1], 0);
    sdp_list_free(apseq, 0);
    sdp_list_free(aproto, 0);
}

void
unregisterBlueproxyService(void) {
    if (sdp_sess != NULL)
        sdp_close(sdp_sess);
}
