/*
 * Copyright (c) 2004 Anil Madhavapeddy <anil@recoil.org>
 * Copyright (c) 2004 Intel Research Cambridge
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $Id: listener.c,v 1.7 2004/09/07 15:52:04 avsm Exp $
 */

/* Simple app that listens for connections on named sockets,
 * and echoes them to stdout */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <fcntl.h>
#include <err.h>
#include <errno.h>
#include <pthread.h>
#include "atomicio.h"
#include "log.h"

const char socket_path[] = "/tmp/bluespot";
const size_t packet_size = 128;

void *echo_socket(void *);
int listen_on_tcp_port(int);

struct env {
    char id[64];
    int sock;
};

int
main(int argc, char **argv) {
    int s, n;
    int id = 0;
    struct sockaddr_in remote_addr;
    socklen_t opt = sizeof remote_addr;
    set_log_level(L_DEBUG);

    logmsg(NULL, L_INFO, "Creating named socket %s", socket_path);

    s = listen_on_tcp_port(20000);

    while (1) {
        struct env *e;
        pthread_t tid;
        logmsg(NULL, L_DEBUG, "waiting for connection...");
        n = accept(s, (struct sockaddr *) &remote_addr, &opt);
        e = malloc(sizeof(struct env));
        if (!e)
            err(1, "malloc");
        if (n == -1)
            err(1, "accept");
        e->sock = n;
        snprintf(e->id, sizeof e->id, "id#%d", id++);
        logmsg(NULL, L_DEBUG, "accepted connection, spawning echo thread");
        pthread_create(&tid, NULL, echo_socket, e);
    }

    return 0;
}

void *
echo_socket(void *s) {
    struct env *e = (struct env *) s;
    char *buf = malloc(packet_size);
    if (!buf) {
        logmsg(e->id, L_CRITICAL, "malloc failed: %s", strerror(errno));
        goto cleanup;
    }
    while (1) {
        ssize_t x;
        memset(buf, 0, packet_size);
        x = atomicio(read, e->sock, buf, packet_size);
        if (x <= 0) {
            logmsg(e->id, L_INFO, "read error (%d): %s", x,
                   strerror(errno));
            goto cleanup;
        }
        logmsg(e->id, L_INFO, "read %d bytes:\n%s", x, buf);
    }

cleanup:
    logmsg(e->id, L_INFO, "terminating thread");
    free(s);
    free(buf);
    return NULL;
}

int
listen_on_tcp_port(int port) {
    int s = socket(AF_INET, SOCK_STREAM, 0);
    struct sockaddr_in sa;

    if (s == -1)
        err(1, "socket");
    memset(&sa, 0, sizeof sa);
    sa.sin_family = AF_INET;
    sa.sin_addr.s_addr = htonl(INADDR_ANY);
    sa.sin_port = htons(port);

    if (bind(s, (struct sockaddr *) &sa, sizeof sa))
        err(1, "bind");
    if (listen(s, 10))
        err(1, "listen");
    return s;
}
